/// Josh's Logger
/// A Redacted Software Product
/// Developed & Maintained by Josh O'Leary
/// This work is dedicated to the public domain.

/// Special thanks: Maxine Hayes, William J Tomasine II

#pragma once

#include <string>
#include <map>
#include <Event.h>
#include <mcolor.h>
#include "token.hpp"
#include "formatter.hpp"
#include "source_location"

namespace jlog {
    /// Severity levels for logging. Higher numbers filter less messages out.
    /// @see LOG_LEVEL macro
    enum class severity : uint8_t {
        none,       // Show no output
        info,       // show only info messages
        fatal,      // Show only fatal errors
        error,      // Show fatal and not-necessarily-fatal errors
        warning,    // Show warnings additionally
        verbose,    // Show errors, warnings, and 'verbose' output (i.e. extra information)
        debug,      // Show debugging messages (Basically everything).
    };

    /// Writes an input string directly to standard output
    /// @note Does not append a newline to the message.
    void log_to_console(const std::string &message);

    /// Writes an input string directly to the specified destination logfile
    /// @note Does not append a newline to the message.
    void log_to_file(const std::string &filename, const std::string &message);


    class Logger {
    public:
        Logger(const std::string& context, const mcolor::ansiColors::Colors& color = mcolor::ansiColors::Colors::FG_DEFAULT);
    public:
        void operator () (const std::string& message);

        virtual void Log(const std::string &message);
        //virtual void Log(const std::vector<token> tokens);
        //virtual void Log(const std::string& context, const std::string &message);
        virtual void LogTrace(const std::string &message, const std::source_location& location = std::source_location::current());
    public:
        virtual void Enabled(bool state);
        virtual bool Enabled();
        virtual void LogFile(const std::string& f);
        virtual std::string LogFile();
        // no cc no bullshit
        virtual void ColorCode(mcolor::ansiColors::Colors cc = mcolor::ansiColors::Colors::FG_DEFAULT);
        virtual mcolor::ansiColors::Colors ColorCode();
    public:
        virtual std::string Context();
        // This whole parent child system is probably overkill for our use cases.
        /**/
        virtual void SetParent(Logger* p);
        virtual Logger* Parent();
        virtual void AddChild(Logger* c);
        virtual void DelChild(const std::string& context);
        virtual bool HasChild(const std::string& context);
        virtual bool HasChild(Logger* c);
        virtual Logger* GetChild(const std::string& context);
        virtual std::map<std::string, Logger*> Children();
        void SetTraceback(bool enabled)
        {
            trace = enabled;
        }
        /**/
    protected:
        bool enabled = true;
        std::string logfile = "latest.log";
        mcolor::ansiColors::Colors colorcode = mcolor::ansiColors::Colors::FG_DEFAULT;
        std::string context;
        Logger* parent;
        std::map<std::string, Logger*> children;
        bool trace = true;
    };
}

namespace jlog {

    Logger INFO{"INFO", mcolor::ansiColors::Colors::FG_GREEN};
    Logger WARNING{"WARNING", mcolor::ansiColors::Colors::FG_YELLOW};
    Logger ERROR{"ERROR", mcolor::ansiColors::Colors::FG_RED};
    Logger FATAL{"FATAL", mcolor::ansiColors::Colors::FG_BRIGHT_RED};
    Logger VERBOSE{"VERBOSE", mcolor::ansiColors::Colors::FG_CYAN};
    Logger DEBUG{"DEBUG", mcolor::ansiColors::Colors::FG_MAGENTA};
}