#include <jlog/jlog.hpp>
#include <fstream>
#include <iostream>
#include <chrono>

/*
#ifdef WIN32
    #define NOMINMAX
    #include <windows.h>
#endif
 */

namespace jlog
{


    void log_to_console(const std::string& message)
    {
#ifdef WIN32
        mcolor::windowsSaneify();
#endif
        std::cout << message;
    }

    void log_to_file(const std::string& filename, const std::string& message)
    {
        std::ofstream latest_log(filename, std::ios_base::app);
        latest_log << message;
        latest_log.close();
    }

    Logger::Logger(const std::string& context, const mcolor::ansiColors::Colors& color) {
        this->context = context;
        this->parent = this;
        this->children = {};
        this->colorcode = color;
    }

    void Logger::Log(const std::string &message) {
        if (!enabled)
            return;

        std::vector<jlog::token> fmt = log_format(this->context, message, this->colorcode);

        if (this == this->parent) {
            jlog::log_to_console(jlog::toks2consoleMsg(fmt) + '\n');
            jlog::log_to_file(this->logfile, jlog::toks2logfileMsg(fmt) + '\n');
        } else {
            // Using toks2consoleMsg is hacky, but it's a simple solution and it's generic enough
            this->parent->Log(jlog::toks2consoleMsg(fmt));
        }
    }

    /*
    void Logger::Log(const std::string& context, const std::string &message) {
        Logger* c = GetChild(context);
        c->Log(message);
    }
     */

    void Logger::LogTrace(const std::string &message, const std::source_location &location) {
        if (!enabled)
            return;

        //std::vector<jlog::token> fmt = log_format(this->context, message, func, file, line,this->colorcode);
        std::vector<jlog::token> fmt = log_format(this->context, message, location.function_name(), location.file_name(), location.line(),this->colorcode);

        if (this == this->parent) {
            jlog::log_to_console(jlog::toks2consoleMsg(fmt) + '\n');
            jlog::log_to_file(this->logfile, jlog::toks2logfileMsg(fmt) + '\n');
        } else {
            // Using toks2consoleMsg is hacky, but it's a simple solution and it's generic enough
            this->parent->Log(jlog::toks2consoleMsg(fmt));
        }
    }

    void Logger::Enabled(bool state) {
        this->enabled = state;
        for (auto const& [_, logger] : this->children) {
            logger->Enabled(state);
        }
    }
    bool Logger::Enabled() { return this->enabled; }

    void Logger::LogFile(const std::string& f) {
        this->logfile = f;
        for (auto const& [_, logger] : this->children) {
            logger->LogFile(f);
        }
    }
    std::string Logger::LogFile() { return this->logfile; }

    void Logger::ColorCode(mcolor::ansiColors::Colors cc) { this->colorcode = cc; }
    mcolor::ansiColors::Colors Logger::ColorCode() { return this->colorcode; }

    std::string Logger::Context() { return this->context; }

    // DOnt be stupid
    void Logger::SetParent(Logger *p) {
        //this->parent->DelChild(this->context);
        this->parent = p;
        // We make the child inherit the parents traits.
        // Is this strictly required? No.
        // Should probably mention this may cause side effects.
        this->Enabled(p->Enabled());
        this->LogFile(p->LogFile());
        //p->AddChild(this);
    }

    Logger* Logger::Parent() { return this->parent; }

    void Logger::AddChild(Logger *c) {
        c->SetParent(this);
        this->children[c->Context()] = c;
    }

    // Child will still keep whatever it previously inherited from the parent
    // C'mon you mean to tell me when your parent disowns you ya gotta give back all
    // the shit you inherited from them? I guess it would be pretty cool to give back
    // debt.
    void Logger::DelChild(const std::string& context) {
        auto c = this->children.find(context);
        if (c != this->children.end()) {
            this->children.erase(context);
            c->second->SetParent(c->second);
        } else {
            throw ("Child does not exist.");
        }
    }

    bool Logger::HasChild(const std::string& context) {
        auto c = this->children.find(context);
        return c != this->children.end();
    }

    bool Logger::HasChild(Logger* c) {
        auto result = std::find_if(
                this->children.begin(), this->children.end(),
                [c](const auto& mo) {return mo.second == c; });
        return result != this->children.end();
    }

    Logger* Logger::GetChild(const std::string& context) {
        if (!HasChild(context)) {
            throw ("Child does not exist.");
        }

        // Yes yes I know we're accessing the map twice making it slower
        // so fuck off this is more flexible.
        auto c = this->children.find(context);

        return c->second;
    }

    std::map<std::string, Logger*> Logger::Children() {
        return this->children;
    }

    void Logger::operator()(const std::string &message) {
        Log(message);
    }

}
